# coding: utf-8
"""
レンダリング画像にReinhardのトーンマッピングをかけます
"""

__author__ = "CASPAR003 <caspar003@caspar003.info>"
__date__ = "2014-07-20/2014-07-20"
__version__ = "1.0"

def ca3_pe_reinhard(img, a, w):
    """
    画像にReinhardのトーンマッピングをかけます
    @param img フィルタをかけたい画像（Shade imageオブジェクト）
    @param a   画像の対数平均輝度の目標値
    @param w   1.0を割り当てる輝度の閾値
    
    参考文献の§3の手法を使用しています．
    画像の対数平均輝度の目標値aは画像全体の明るさを制御するパラメータです．元の論文には，「0.18を基準に0.045～0.72くらいの範囲で選べば良い」と書かれています．
    1.0を割り当てる輝度の閾値はwは正の実数で，例として3.0を指定した場合は，リニアにマッピングする場合の3倍の輝度までを圧縮してマッピングするするという意味になります（元論文のFig.6を見てください）．値は3.0くらいが良さそうです．元の論文では「とりあえず画像の最大輝度を使う」と書かれていますが，光源が直接写り込むシーン等では適当でない気がします．
    
    参考文献
        Erik Reinhard, Mike Stark, Peter Shirley and Jim Ferwerda. "Photographic Tone Reproduction for Digital Images".
        URI: http://www.cs.utah.edu/~reinhard/cdrom/
    """
    import math
    import numpy as np
    
    # 作業用配列の確保
    matr = np.zeros(img.size)
    matg = np.zeros(img.size)
    matb = np.zeros(img.size)
    
    # Shade imageをnumpy arrayに代入
    for ix in range(img.size[0]):
        for iy in range(img.size[1]):
            matr[ix, iy], matg[ix, iy], matb[ix, iy] = img.get_pixel(ix, iy)
    
    # 輝度値Y （sRGB→Y）
    matY = 0.212639*matr + 0.715169*matg + 0.072192*matb
    
    # 対数平均輝度Lw
    matW = matY.copy()
    matW = matW + 1.0e-6               # eq.(1)
    matW = np.log(matW)                # 
    Lw = matW.sum()                    # 
    Lw /= img.size[0] * img.size[1]    # 
    Lw = math.exp(Lw)                  # 
    del matW
    
    # 輝度のスケーリング
    matL = (a/Lw) * matY    # eq.(2)
    
    # トーンマッピング
    matLd = matL * (1.0 + matL / w**2) / (1.0 + matL)    # eq.(4)
    del matL
    
    # トーンマップを適用
    matS = matLd / matY
    matr = matS * matr
    matg = matS * matg
    matb = matS * matb
    
    # numpy arrayをShade imageに戻す
    for ix in range(img.size[0]):
        for iy in range(img.size[1]):
            img.set_pixel(ix, iy, (matr[ix, iy], matg[ix, iy], matb[ix, iy]))
    return



if (__name__ == "__main__"):
    img = xshade.scene().rendering.image
    if (img is not None):
        ca3_pe_reinhard(img, 0.045, 3.0)
        xshade.scene().rendering.update_image()
